@@ -16,18 +16,12 @@
 
 from __future__ import absolute_import, print_function
 
-import datetime
 import json
 import logging
 import os
-import shlex
-import subprocess as sub
 import sys
-import tarfile
 import time
 from builtins import str
-from shutil import rmtree
-from tempfile import mkdtemp, mkstemp
 
 import boto3
 import pkg_resources
@@ -38,11 +32,6 @@ import pcluster.utils as utils
 from pcluster.config.pcluster_config import PclusterConfig
 from pcluster.constants import PCLUSTER_STACK_PREFIX
 
-if sys.version_info[0] >= 3:
-    from urllib.request import urlretrieve
-else:
-    from urllib import urlretrieve  # pylint: disable=no-name-in-module
-
 LOGGER = logging.getLogger(__name__)
 
 
@@ -241,69 +230,6 @@ def _is_ganglia_enabled(parameters):
     return is_ganglia_enabled
 
 
-def update(args):  # noqa: C901 FIXME!!!
-    LOGGER.info("Updating: %s", args.cluster_name)
-    stack_name = utils.get_stack_name(args.cluster_name)
-    pcluster_config = PclusterConfig(
-        config_file=args.config_file, cluster_label=args.cluster_template, fail_on_file_absence=True
-    )
-    pcluster_config.validate()
-    cfn_params = pcluster_config.to_cfn()
-
-    cluster_section = pcluster_config.get_section("cluster")
-    cfn = boto3.client("cloudformation")
-    if cluster_section.get_param_value("scheduler") != "awsbatch":
-        if not args.reset_desired:
-            asg_name = _get_asg_name(stack_name)
-            desired_capacity = (
-                boto3.client("autoscaling")
-                .describe_auto_scaling_groups(AutoScalingGroupNames=[asg_name])
-                .get("AutoScalingGroups")[0]
-                .get("DesiredCapacity")
-            )
-            cfn_params["DesiredSize"] = str(desired_capacity)
-    else:
-        if args.reset_desired:
-            LOGGER.info("reset_desired flag does not work with awsbatch scheduler")
-        params = utils.get_stack(stack_name, cfn).get("Parameters")
-
-        for parameter in params:
-            if parameter.get("ParameterKey") == "ResourcesS3Bucket":
-                cfn_params["ResourcesS3Bucket"] = parameter.get("ParameterValue")
-
-    try:
-        LOGGER.debug(cfn_params)
-        if args.extra_parameters:
-            LOGGER.debug("Adding extra parameters to the CFN parameters")
-            cfn_params.update(dict(args.extra_parameters))
-
-        cfn_params = [{"ParameterKey": key, "ParameterValue": value} for key, value in cfn_params.items()]
-        LOGGER.info("Calling update_stack")
-        cfn.update_stack(
-            StackName=stack_name, UsePreviousTemplate=True, Parameters=cfn_params, Capabilities=["CAPABILITY_IAM"]
-        )
-        stack_status = utils.get_stack(stack_name, cfn).get("StackStatus")
-        if not args.nowait:
-            while stack_status == "UPDATE_IN_PROGRESS":
-                stack_status = utils.get_stack(stack_name, cfn).get("StackStatus")
-                events = utils.get_stack_events(stack_name)[0]
-                resource_status = (
-                    "Status: %s - %s" % (events.get("LogicalResourceId"), events.get("ResourceStatus"))
-                ).ljust(80)
-                sys.stdout.write("\r%s" % resource_status)
-                sys.stdout.flush()
-                time.sleep(5)
-        else:
-            stack_status = utils.get_stack(stack_name, cfn).get("StackStatus")
-            LOGGER.info("Status: %s", stack_status)
-    except ClientError as e:
-        LOGGER.critical(e.response.get("Error").get("Message"))
-        sys.exit(1)
-    except KeyboardInterrupt:
-        LOGGER.info("\nExiting...")
-        sys.exit(0)
-
-
 def start(args):
     """Restore ASG limits or awsbatch CE to min/max/desired."""
     stack_name = utils.get_stack_name(args.cluster_name)
@@ -315,7 +241,7 @@ def start(args):
         max_vcpus = cluster_section.get_param_value("max_vcpus")
         desired_vcpus = cluster_section.get_param_value("desired_vcpus")
         min_vcpus = cluster_section.get_param_value("min_vcpus")
-        ce_name = _get_batch_ce(stack_name)
+        ce_name = utils.get_batch_ce(stack_name)
         _start_batch_ce(ce_name=ce_name, min_vcpus=min_vcpus, desired_vcpus=desired_vcpus, max_vcpus=max_vcpus)
     else:
         LOGGER.info("Starting compute fleet : %s", args.cluster_name)
@@ -325,8 +251,8 @@ def start(args):
             if cluster_section.get_param_value("maintain_initial_size")
             else 0
         )
-        asg_name = _get_asg_name(stack_name)
-        _set_asg_limits(asg_name=asg_name, min=min_desired_size, max=max_queue_size, desired=min_desired_size)
+        asg_name = utils.get_asg_name(stack_name)
+        utils.set_asg_limits(asg_name=asg_name, min=min_desired_size, max=max_queue_size, desired=min_desired_size)
 
 
 def stop(args):
@@ -337,24 +263,12 @@ def stop(args):
 
     if cluster_section.get_param_value("scheduler") == "awsbatch":
         LOGGER.info("Disabling AWS Batch compute environment : %s", args.cluster_name)
-        ce_name = _get_batch_ce(stack_name)
+        ce_name = utils.get_batch_ce(stack_name)
         _stop_batch_ce(ce_name=ce_name)
     else:
         LOGGER.info("Stopping compute fleet : %s", args.cluster_name)
-        asg_name = _get_asg_name(stack_name)
-        _set_asg_limits(asg_name=asg_name, min=0, max=0, desired=0)
-
-
-def _get_batch_ce(stack_name):
-    """
-    Get name of the AWS Batch Compute Environment.
-
-    :param stack_name: name of the master stack
-    :param config: config
-    :return: ce_name or exit if not found
-    """
-    outputs = utils.get_stack(stack_name).get("Outputs")
-    return utils.get_stack_output_value(outputs, "BatchComputeEnvironmentArn")
+        asg_name = utils.get_asg_name(stack_name)
+        utils.set_asg_limits(asg_name=asg_name, min=0, max=0, desired=0)
 
 
 def _get_pcluster_version_from_stack(stack):
@@ -456,39 +370,6 @@ def _get_ec2_instances(stack):
     return stack_instances
 
 
-def _get_asg_name(stack_name):
-    try:
-        resources = utils.get_stack_resources(stack_name)
-        return [r for r in resources if r.get("LogicalResourceId") == "ComputeFleet"][0].get("PhysicalResourceId")
-    except IndexError:
-        LOGGER.critical("Stack %s does not have a ComputeFleet", stack_name)
-        sys.exit(1)
-
-
-def _set_asg_limits(asg_name, min, max, desired):
-    asg = boto3.client("autoscaling")
-    asg.update_auto_scaling_group(
-        AutoScalingGroupName=asg_name, MinSize=int(min), MaxSize=int(max), DesiredCapacity=int(desired)
-    )
-
-
-def _get_asg_instances(stack):
-    asg = boto3.client("autoscaling")
-    asg_name = _get_asg_name(stack)
-    auto_scaling_groups = asg.describe_auto_scaling_groups(AutoScalingGroupNames=[asg_name]).get("AutoScalingGroups")
-    if not auto_scaling_groups:
-        LOGGER.error("Unable to retrieve ASG info. Please check cluster status.")
-        sys.exit(1)
-    asg = auto_scaling_groups[0]
-    name = [tag.get("Value") for tag in asg.get("Tags") if tag.get("Key") == "aws:cloudformation:logical-id"][0]
-
-    temp_instances = []
-    for instance in asg.get("Instances"):
-        temp_instances.append([name, instance.get("InstanceId")])
-
-    return temp_instances
-
-
 def _start_batch_ce(ce_name, min_vcpus, desired_vcpus, max_vcpus):
     try:
         boto3.client("batch").update_compute_environment(
@@ -518,7 +399,7 @@ def instances(args):
     instances.extend(_get_ec2_instances(stack_name))
 
     if cluster_section.get_param_value("scheduler") != "awsbatch":
-        instances.extend(_get_asg_instances(stack_name))
+        instances.extend(utils.get_asg_instances(stack_name))
 
     for instance in instances:
         LOGGER.info("%s         %s", instance[0], instance[1])
@@ -641,20 +522,27 @@ def _persist_stack_resources(stack, template, keys):
     """Set the resources in template identified by keys to have a DeletionPolicy of 'Retain'."""
     for key in keys:
         template["Resources"][key]["DeletionPolicy"] = "Retain"
-    utils.update_stack_template(stack.get("StackName"), template, stack.get("Parameters"))
+    try:
+        utils.update_stack_template(stack.get("StackName"), template, stack.get("Parameters"))
+    except ClientError as client_err:
+        utils.error(
+            "Unable to persist logs on cluster deletion, failed with error: {emsg}.\n"
+            "If you want to continue, please retry without the --keep-logs flag.".format(
+                emsg=client_err.response.get("Error").get("Message")
+            )
+        )
 
 
 def _persist_cloudwatch_log_groups(cluster_name):
     """Enable cluster's CloudWatch log groups to persist past cluster deletion."""
     LOGGER.info("Configuring {0}'s CloudWatch log groups to persist past cluster deletion.".format(cluster_name))
     substacks = utils.get_cluster_substacks(cluster_name)
-    substack_template_pairs = [(stack, utils.get_stack_template(stack.get("StackName"))) for stack in substacks]
-    substack_template_keys_triplets = [
-        (s, t, _get_unretained_cw_log_group_resource_keys(t)) for s, t in substack_template_pairs
-    ]
-    for stack, template, keys in substack_template_keys_triplets:
-        if keys:  # Only persist the CloudWatch group
-            _persist_stack_resources(stack, template, keys)
+    cw_substack = next((stack for stack in substacks if "CloudWatchLogsSubstack" in stack.get("StackName")), None)
+    if cw_substack:
+        cw_substack_template = utils.get_stack_template(cw_substack.get("StackName"))
+        log_group_keys = _get_unretained_cw_log_group_resource_keys(cw_substack_template)
+        if log_group_keys:  # Only persist the CloudWatch group
+            _persist_stack_resources(cw_substack, cw_substack_template, log_group_keys)
 
 
 def _delete_cluster(cluster_name, nowait):
@@ -721,212 +609,9 @@ def delete(args):
     _delete_cluster(args.cluster_name, args.nowait)
 
 
-def _get_cookbook_url(region, template_url, args, tmpdir):
-    if args.custom_ami_cookbook is not None:
-        return args.custom_ami_cookbook
-
-    cookbook_version = _get_cookbook_version(template_url, tmpdir)
-    s3_suffix = ".cn" if region.startswith("cn") else ""
-    return "https://s3.%s.amazonaws.com%s/%s-aws-parallelcluster/cookbooks/%s.tgz" % (
-        region,
-        s3_suffix,
-        region,
-        cookbook_version,
-    )
-
-
-def _get_cookbook_version(template_url, tmpdir):
-    tmp_template_file = os.path.join(tmpdir, "aws-parallelcluster-template.json")
-    try:
-        LOGGER.info("Template: %s", template_url)
-        urlretrieve(url=template_url, filename=tmp_template_file)
-
-        with open(tmp_template_file) as cfn_file:
-            cfn_data = json.load(cfn_file)
-
-        return cfn_data.get("Mappings").get("PackagesVersions").get("default").get("cookbook")
-
-    except IOError as e:
-        LOGGER.error("Unable to download template at URL %s", template_url)
-        LOGGER.critical("Error: %s", str(e))
-        sys.exit(1)
-    except (ValueError, AttributeError) as e:
-        LOGGER.error("Unable to parse template at URL %s", template_url)
-        LOGGER.critical("Error: %s", str(e))
-        sys.exit(1)
-
-
-def _get_cookbook_dir(region, template_url, args, tmpdir):
-    cookbook_url = ""
-    try:
-        tmp_cookbook_archive = os.path.join(tmpdir, "aws-parallelcluster-cookbook.tgz")
-
-        cookbook_url = _get_cookbook_url(region, template_url, args, tmpdir)
-        LOGGER.info("Cookbook: %s", cookbook_url)
-
-        urlretrieve(url=cookbook_url, filename=tmp_cookbook_archive)
-        tar = tarfile.open(tmp_cookbook_archive)
-        cookbook_archive_root = tar.firstmember.path
-        tar.extractall(path=tmpdir)
-        tar.close()
-
-        return os.path.join(tmpdir, cookbook_archive_root)
-    except (IOError, tarfile.ReadError) as e:
-        LOGGER.error("Unable to download cookbook at URL %s", cookbook_url)
-        LOGGER.critical("Error: %s", str(e))
-        sys.exit(1)
-
-
-def _dispose_packer_instance(results):
-    time.sleep(2)
-    try:
-        ec2_client = boto3.client("ec2")
-        instance = ec2_client.describe_instance_status(
-            InstanceIds=[results["PACKER_INSTANCE_ID"]], IncludeAllInstances=True
-        ).get("InstanceStatuses")[0]
-        instance_state = instance.get("InstanceState").get("Name")
-        if instance_state in ["running", "pending", "stopping", "stopped"]:
-            LOGGER.info("Terminating Instance %s created by Packer", results["PACKER_INSTANCE_ID"])
-            ec2_client.terminate_instances(InstanceIds=[results["PACKER_INSTANCE_ID"]])
-
-    except ClientError as e:
-        LOGGER.critical(e.response.get("Error").get("Message"))
-        sys.exit(1)
-
-
-def _run_packer(packer_command, packer_env):
-    erase_line = "\x1b[2K"
-    _command = shlex.split(packer_command)
-    results = {}
-    _, path_log = mkstemp(prefix="packer.log." + datetime.datetime.now().strftime("%Y%m%d-%H%M%S" + "."), text=True)
-    LOGGER.info("Packer log: %s", path_log)
-    try:
-        dev_null = open(os.devnull, "rb")
-        packer_env.update(os.environ.copy())
-        process = sub.Popen(
-            _command, env=packer_env, stdout=sub.PIPE, stderr=sub.STDOUT, stdin=dev_null, universal_newlines=True
-        )
-
-        with open(path_log, "w") as packer_log:
-            while process.poll() is None:
-                output_line = process.stdout.readline().strip()
-                packer_log.write("\n%s" % output_line)
-                packer_log.flush()
-                sys.stdout.write(erase_line)
-                sys.stdout.write("\rPacker status: %s" % output_line[:90] + (output_line[90:] and ".."))
-                sys.stdout.flush()
-
-                if output_line.find("packer build") > 0:
-                    results["PACKER_COMMAND"] = output_line
-                if output_line.find("Instance ID:") > 0:
-                    results["PACKER_INSTANCE_ID"] = output_line.rsplit(":", 1)[1].strip(" \n\t")
-                    sys.stdout.write(erase_line)
-                    sys.stdout.write("\rPacker Instance ID: %s\n" % results["PACKER_INSTANCE_ID"])
-                    sys.stdout.flush()
-                if output_line.find("AMI:") > 0:
-                    results["PACKER_CREATED_AMI"] = output_line.rsplit(":", 1)[1].strip(" \n\t")
-                if output_line.find("Prevalidating AMI Name:") > 0:
-                    results["PACKER_CREATED_AMI_NAME"] = output_line.rsplit(":", 1)[1].strip(" \n\t")
-        sys.stdout.write("\texit code %s\n" % process.returncode)
-        sys.stdout.flush()
-        return results
-    except sub.CalledProcessError:
-        sys.stdout.flush()
-        LOGGER.error("Failed to run %s\n", _command)
-        sys.exit(1)
-    except (IOError, OSError):  # noqa: B014
-        sys.stdout.flush()
-        LOGGER.error("Failed to run %s\nCommand not found", packer_command)
-        sys.exit(1)
-    except KeyboardInterrupt:
-        sys.stdout.flush()
-        LOGGER.info("\nExiting...")
-        sys.exit(0)
-    finally:
-        dev_null.close()
-        if results.get("PACKER_INSTANCE_ID"):
-            _dispose_packer_instance(results)
-
-
-def _print_create_ami_results(results):
-    if results.get("PACKER_CREATED_AMI"):
-        LOGGER.info(
-            "\nCustom AMI %s created with name %s", results["PACKER_CREATED_AMI"], results["PACKER_CREATED_AMI_NAME"]
-        )
-        print(
-            "\nTo use it, add the following variable to the AWS ParallelCluster config file, "
-            "under the [cluster ...] section"
-        )
-        print("custom_ami = %s" % results["PACKER_CREATED_AMI"])
-    else:
-        LOGGER.info("\nNo custom AMI created")
-
-
-def create_ami(args):
-    LOGGER.info("Building AWS ParallelCluster AMI. This could take a while...")
-    LOGGER.debug("Building AMI based on args %s", str(args))
-    results = {}
-
-    instance_type = args.instance_type
-    try:
-        # FIXME it doesn't work if there is no a default section
-        pcluster_config = PclusterConfig(config_file=args.config_file, fail_on_file_absence=True)
-
-        vpc_section = pcluster_config.get_section("vpc")
-        vpc_id = args.vpc_id if args.vpc_id else vpc_section.get_param_value("vpc_id")
-        subnet_id = args.subnet_id if args.subnet_id else vpc_section.get_param_value("master_subnet_id")
-
-        packer_env = {
-            "CUSTOM_AMI_ID": args.base_ami_id,
-            "AWS_FLAVOR_ID": instance_type,
-            "AMI_NAME_PREFIX": args.custom_ami_name_prefix,
-            "AWS_VPC_ID": vpc_id,
-            "AWS_SUBNET_ID": subnet_id,
-            "ASSOCIATE_PUBLIC_IP": "true" if args.associate_public_ip else "false",
-        }
-
-        aws_section = pcluster_config.get_section("aws")
-        aws_region = aws_section.get_param_value("aws_region_name")
-        if aws_section and aws_section.get_param_value("aws_access_key_id"):
-            packer_env["AWS_ACCESS_KEY_ID"] = aws_section.get_param_value("aws_access_key_id")
-        if aws_section and aws_section.get_param_value("aws_secret_access_key"):
-            packer_env["AWS_SECRET_ACCESS_KEY"] = aws_section.get_param_value("aws_secret_access_key")
-
-        LOGGER.info("Base AMI ID: %s", args.base_ami_id)
-        LOGGER.info("Base AMI OS: %s", args.base_ami_os)
-        LOGGER.info("Instance Type: %s", instance_type)
-        LOGGER.info("Region: %s", aws_region)
-        LOGGER.info("VPC ID: %s", vpc_id)
-        LOGGER.info("Subnet ID: %s", subnet_id)
-
-        template_url = _evaluate_pcluster_template_url(pcluster_config)
-
-        tmp_dir = mkdtemp()
-        cookbook_dir = _get_cookbook_dir(aws_region, template_url, args, tmp_dir)
-
-        packer_command = (
-            cookbook_dir
-            + "/amis/build_ami.sh --os "
-            + args.base_ami_os
-            + " --partition region"
-            + " --region "
-            + aws_region
-            + " --custom"
-        )
-
-        results = _run_packer(packer_command, packer_env)
-    except KeyboardInterrupt:
-        LOGGER.info("\nExiting...")
-        sys.exit(0)
-    finally:
-        _print_create_ami_results(results)
-        if "tmp_dir" in locals() and tmp_dir:
-            rmtree(tmp_dir)
-
-
 def _get_default_template_url(region):
     return (
-        "https://s3.{REGION}.amazonaws.com{SUFFIX}/{REGION}-aws-parallelcluster/templates/"
+        "https://{REGION}-aws-parallelcluster.s3.{REGION}.amazonaws.com{SUFFIX}/templates/"
         "aws-parallelcluster-{VERSION}.cfn.json".format(
             REGION=region, SUFFIX=".cn" if region.startswith("cn") else "", VERSION=utils.get_installed_version()
         )
